/**
 * Copyright (c) 2015-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.descriptor;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * 
 * @author Victor Noël - Linagora
 *
 * @param <JbiType>
 */
public class AbstractJBIDescriptorBuilder<JbiType> extends AbstractJAXBSingleton {

    /**
     * Standard location to the JBI Descriptor resource.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF" + File.separator + "jbi.xml";

    /**
     * Standard location to the JBI Descriptor resource into an archive.
     */
    public static final String JBI_DESCRIPTOR_RESOURCE_IN_ARCHIVE = "META-INF/jbi.xml";

    private final Class<JbiType> jbiClazz;

    public AbstractJBIDescriptorBuilder(final Class<JbiType> jbiClazz) {
        this.jbiClazz = jbiClazz;
    }

    /**
     * Validate the given JBI descriptor against the JBI schema.
     * 
     * @param jbiDescriptorStream
     * @throws JBIDescriptorException
     *             The validation has failed
     */
    public void validateJBIDescriptor(final InputStream jbiDescriptorStream) throws JBIDescriptorException {
        try {
            // The validator is not thread safe
            synchronized (getValidator()) {
                getValidator().validate(new StreamSource(jbiDescriptorStream));
            }
        } catch (final SAXException | IOException e) {
            throw new JBIDescriptorException("Failed to validate JBI descriptor against JBI schema", e);
        }
    }

    /**
     * Build the Java bindings from a JBI descriptor of an exploded archive
     * 
     * @param folder
     *            The directory containing the exploded archive
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     */
    public JbiType buildJavaJBIDescriptorFromFolder(final File folder) throws JBIDescriptorException {

        final File tmpJBIFile = new File(folder, JBI_DESCRIPTOR_RESOURCE);
        if (!tmpJBIFile.exists()) {
            throw new JBIDescriptorException("JBI descriptor not found in the package '" + folder.getAbsolutePath()
                    + "'");
        }

        try (final InputStream tmpJBIFileAsInputStream = new FileInputStream(tmpJBIFile)) {
            return buildJavaJBIDescriptor(tmpJBIFileAsInputStream);
        } catch (final FileNotFoundException e) {
            // it does not actually check if the file exists...
            throw new JBIDescriptorException("JBI descriptor not found in the package '" + folder.getAbsolutePath()
                    + "'");
        } catch (final IOException e) {
            throw new JBIDescriptorException("Can't load JBI descriptor from the package '" + folder.getAbsolutePath()
                    + "'", e);
        }
    }

    /**
     * Build the Java bindings from the JBI descriptor in the given archive as {@link File}
     * 
     * @param archive
     *            An archive in which the JBI descriptor to read is located
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     */
    public JbiType buildJavaJBIDescriptorFromArchive(final File archive) throws JBIDescriptorException {
        try (final ZipFile zipArchive = new ZipFile(archive)) {
            return buildJavaJBIDescriptorFromArchive(zipArchive);
        } catch (final IOException e) {
            throw new JBIDescriptorException("Can't load JBI descriptor from the archive '" + archive.getAbsolutePath()
                    + "'", e);
        }
    }

    /**
     * Build the Java bindings from the JBI descriptor in the given archive as {@link ZipFile}
     * 
     * @param zipArchive
     *            An archive in which the JBI descriptor to read is located
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     */
    public JbiType buildJavaJBIDescriptorFromArchive(final ZipFile zipArchive) throws JBIDescriptorException {
        try {
            final ZipEntry entry = zipArchive.getEntry(JBI_DESCRIPTOR_RESOURCE_IN_ARCHIVE);
            if (entry == null) {
                throw new JBIDescriptorException("JBI descriptor not found in the archive '" + zipArchive.getName()
                        + "'");
            }
            try (final InputStream tmpJBIFileAsInputStream = zipArchive.getInputStream(entry)) {
                return buildJavaJBIDescriptor(tmpJBIFileAsInputStream);
            }
        } catch (final IOException e) {
            throw new JBIDescriptorException("Can't load JBI descriptor from the archive '" + zipArchive.getName()
                    + "'", e);
        }
    }

    /**
     * <p>
     * Build the Java bindings from the JBI descriptor in XML form. The validation against the JBI schema is processed
     * during the transformation.
     * </p>
     * <p>
     * <b>Note</b>: The input stream is closed because of the XML parser.
     * </p>
     * 
     * @param jbiDescriptor
     *            The JBI descriptor in XML form as {@link Source}
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     *             The exception raised during the unmarshaller creation or the exception raised during the build of the
     *             java bindings.
     */
    public JbiType buildJavaJBIDescriptor(final Source jbiDescriptor) throws JBIDescriptorException {

        try {
            final JAXBElement<JbiType> jbiBinding;

            // The default Xerces unmarshaller is not thread safe
            synchronized (this.getUnmarshaller()) {
                jbiBinding = this.getUnmarshaller().unmarshal(jbiDescriptor, jbiClazz);
            }

            return jbiBinding.getValue();

        } catch (final JAXBException e) {
            throw new JBIDescriptorException("Failed to build Java bindings from JBI descriptor XML document", e);
        }
    }

    /**
     * <p>
     * Build the Java bindings from the JBI descriptor in XML form. The validation against the JBI schema is processed
     * during the transformation.
     * </p>
     * <p>
     * <b>Note</b>: The input stream is closed because of the XML parser.
     * </p>
     * 
     * @param jbiDescriptorStream
     *            The JBI descriptor in XML form as {@link InputStream}
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     *             The exception raised during the unmarshaller creation or the exception raised during the build of the
     *             java bindings.
     */
    public JbiType buildJavaJBIDescriptor(final InputStream jbiDescriptorStream) throws JBIDescriptorException {
        return this.buildJavaJBIDescriptor(new StreamSource(jbiDescriptorStream));
    }

    /**
     * <p>
     * Build the Java bindings from the JBI descriptor in XML form. The validation against the JBI schema is processed
     * during the transformation.
     * </p>
     * <p>
     * <b>Note</b>: The input stream is closed because of the XML parser.
     * </p>
     * 
     * @param jbiDescriptor
     *            The JBI descriptor in XML form as {@link Reader}
     * @return The root object (instance of {@link JbiType}) of the JBI descriptor
     * @throws JBIDescriptorException
     *             The exception raised during the unmarshaller creation or the exception raised during the build of the
     *             java bindings.
     */
    public JbiType buildJavaJBIDescriptor(final Reader jbiDescriptor) throws JBIDescriptorException {
        return this.buildJavaJBIDescriptor(new StreamSource(jbiDescriptor));
    }

    /**
     * Build the XML nodes from the JBI descriptor in Java classes form.
     * 
     * @param jbiDescriptorClass
     *            The JBI Descriptor root class
     * @param jbiDescriptorNode
     *            The XML Node to fill with the JBI descriptor XML nodes
     * @throws JBIDescriptorException
     *             The exception raised during the unmarshaller creation or the exception raised during the build of the
     *             XML nodes.
     */
    public void writeXmlJBIdescriptor(final JbiType jbiDescriptorClass, final Node jbiDescriptorNode)
            throws JBIDescriptorException {

        try {
            // The default Xerces marshaller is not thread safe
            synchronized (this.getMarshaller()) {
                this.getMarshaller().marshal(jbiDescriptorClass, jbiDescriptorNode);
            }
        } catch (final JAXBException e) {
            throw new JBIDescriptorException("Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }

    /**
     * <p>
     * Write the JBI descriptor XML into an output stream.
     * </p>
     * <p>
     * <b>Note</b>: The output stream is not close by this method.
     * </p>
     * 
     * @param jbiDescriptorClass
     *            The JBI Descriptor root class
     * @param jbiDescrOutputStream
     *            The output stream where the JBI descriptor will be written
     * @throws JBIDescriptorException
     *             The exception raised during the marshaller creation or the exception raised during the build of the
     *             XML string.
     */
    public void writeXMLJBIdescriptor(final JbiType jbiDescriptorClass, final OutputStream jbiDescrOutputStream)
            throws JBIDescriptorException {

        try {
            // The default Xerces marshaller is not thread safe
            synchronized (this.getMarshaller()) {
                this.getMarshaller().marshal(jbiDescriptorClass, jbiDescrOutputStream);
            }
        } catch (final JAXBException e) {
            throw new JBIDescriptorException("Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }
}
